import pymafia.kolmafia as km
from pymafia import ash, datatypes


class Familiar:
    id = -1
    name = "none"

    def __init__(self, key=None):
        if key in (None, self.id, self.name):
            return

        id_ = km.FamiliarDatabase.getFamiliarId(key) if isinstance(key, str) else key
        name = km.FamiliarDatabase.getFamiliarName(id_)

        if name is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        self.id = id_
        self.name = name

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"{type(self).__name__}({str(self)!r})"

    def __hash__(self):
        return hash((self.id, self.name))

    def __eq__(self, other):
        return isinstance(other, type(self)) and (self.id, self.name) == (
            other.id,
            other.name,
        )

    def __bool__(self):
        return (self.id, self.name) != (type(self).id, type(self).name)

    @classmethod
    def all(cls):
        values = km.DataTypes.FAMILIAR_TYPE.allValues()
        return sorted(ash.to_python(values), key=lambda x: x.id)

    @property
    def hatchling(self):
        hatchling_id = km.FamiliarDatabase.getFamiliarLarva(self.id)
        return datatypes.Item(hatchling_id) or None

    @property
    def image(self):
        return km.FamiliarDatabase.getFamiliarImageLocation(self.id)

    @property
    def nickname(self):
        fam = km.KoLCharacter.findFamiliar(self.id)
        return fam.getName()

    @property
    def experience(self):
        fam = km.KoLCharacter.findFamiliar(self.id)
        return fam.getTotalExperience()

    @property
    def charges(self):
        fam = km.KoLCharacter.findFamiliar(self.id)
        return fam.getCharges()

    @property
    def drop_name(self):
        return km.FamiliarData.dropName(self.id) or ""

    @property
    def drop_item(self):
        item = km.FamiliarData.dropItem(self.id)
        return None if item is None else datatypes.Item(item.getItemId())

    @property
    def drops_today(self):
        return km.FamiliarData.dropsToday(self.id)

    @property
    def drops_limit(self):
        return km.FamiliarData.dropsToday(self.id)

    @property
    def fights_today(self):
        return km.FamiliarData.fightsToday(self.id)

    @property
    def fights_limit(self):
        return km.FamiliarData.fightDailyCap(self.id)

    @property
    def combat(self):
        return km.FamiliarDatabase.isCombatType(self.id)

    @property
    def physical_damage(self):
        return km.FamiliarDatabase.isCombat0Type(self.id)

    @property
    def elemental_damage(self):
        return km.FamiliarDatabase.isCombat1Type(self.id)

    @property
    def block(self):
        return km.FamiliarDatabase.isBlockType(self.id)

    @property
    def delevel(self):
        return km.FamiliarDatabase.isDelevelType(self.id)

    @property
    def hp_during_combat(self):
        return km.FamiliarDatabase.isHp0Type(self.id)

    @property
    def mp_during_combat(self):
        return km.FamiliarDatabase.isMp0Type(self.id)

    @property
    def other_action_during_combat(self):
        return km.FamiliarDatabase.isOther0Type(self.id)

    @property
    def hp_after_combat(self):
        return km.FamiliarDatabase.isHp1Type(self.id)

    @property
    def mp_after_combat(self):
        return km.FamiliarDatabase.isMp1Type(self.id)

    @property
    def other_action_after_combat(self):
        return km.FamiliarDatabase.isOther1Type(self.id)

    @property
    def passive(self):
        return km.FamiliarDatabase.isPassiveType(self.id)

    @property
    def underwater(self):
        return km.FamiliarDatabase.isUnderwaterType(self.id)

    @property
    def variable(self):
        return km.FamiliarDatabase.isVariableType(self.id)

    @property
    def attributes(self):
        attrs = km.FamiliarDatabase.getFamiliarAttributes(self.id)
        return [] if attrs is None else list(attrs)

    @property
    def poke_level(self):
        fam = km.KoLCharacter.findFamiliar(self.id)
        return 0 if fam is None else fam.getPokeLevel()

    @property
    def poke_level_2_power(self):
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return 0 if data is None else data.getPower2()

    @property
    def poke_level_2_hp(self):
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return 0 if data is None else data.getHP2()

    @property
    def poke_level_3_power(self):
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return 0 if data is None else data.getPower3()

    @property
    def poke_level_3_hp(self):
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return 0 if data is None else data.getHP3()

    @property
    def poke_level_4_power(self):
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return 0 if data is None else data.getPower4()

    @property
    def poke_level_4_hp(self):
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return 0 if data is None else data.getHP4()

    @property
    def poke_move_1(self):
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return "" if data is None else data.getMove1()

    @property
    def poke_move_2(self):
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return "" if data is None else data.getMove2()

    @property
    def poke_move_3(self):
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return "" if data is None else data.getMove3()

    @property
    def poke_attribute(self):
        data = km.FamiliarDatabase.getPokeDataById(self.id)
        return "" if data is None else data.getAttribute()



from pymafia import ash
from pymafia.combat import Macro
from pymafia.datatypes import Effect, Item, Location
from pymafia.utils import get_property, set_property

equipment_choices = {
    Item("LOV Eardigan"): 1,
    Item("LOV Epaulettes"): 2,
    Item("LOV Earring"): 3,
}
effect_choices = {
    Effect("Lovebotamy"): 1,
    Effect("Open Heart Surgery"): 2,
    Effect("Wandering Eye Surgery"): 3,
}
item_choices = {
    Item("LOV Enamorang"): 1,
    Item("LOV Emotionizer"): 2,
    Item("LOV Extraterrestrial Chocolate"): 3,
    Item("LOV Echinacea Bouquet"): 4,
    Item("LOV Elephant"): 5,
    Item("toast"): 6,
    None: 7,
}


def have():
    """Return True if the player has The Tunnel of L.O.V.E available, False otherwise."""
    return get_property("loveTunnelAvailable", bool)


def is_used():
    """Return True if The Tunnel of L.O.V.E has been used today, False otherwise."""
    return get_property("_loveTunnelUsed", bool)


def fight_all(equipment, effect, item, macro=Macro()):
    """Fight all LOV monsters and get buffs/equipment."""
    if not have():
        raise RuntimeError("need access to The Tunnel of L.O.V.E")
    if is_used():
        raise RuntimeError("love tunnel has already been used today")

    set_property("choiceAdventure1222", 1)  # Entrance
    set_property("choiceAdventure1223", 1)  # Fight LOV Enforcer
    set_property("choiceAdventure1224", equipment_choices[equipment])
    set_property("choiceAdventure1225", 1)  # Fight LOV Engineer
    set_property("choiceAdventure1226", effect_choices[effect])
    set_property("choiceAdventure1227", 1)  # Fight LOV Equivocator
    set_property("choiceAdventure1228", item_choices[item])
    success = ash.adv1(Location("The Tunnel of L.O.V.E."), -1, macro)

    if not success:
        raise RuntimeError("failed to use love tunnel")



from pymafia import ash
from pymafia.datatypes import Element, Familiar, Item, Stat
from pymafia.utils import get_property
from pymafia.utils import have as _have

familiar = Familiar("Crimbo Shrub")
decorations = Item("box of old Crimbo decorations")

topper_choices = {Stat.MUSCLE: 1, Stat.MYSTICALITY: 2, Stat.MOXIE: 3}
lights_choices = {
    "prismatic": 1,
    Element.HOT: 2,
    Element.COLD: 3,
    Element.STENCH: 4,
    Element.SPOOKY: 5,
    Element.SLEAZE: 6,
}
garland_choices = {"HP": 1, "PvP": 2, "blocking": 3}
gift_choices = {"yellow": 1, "meat": 2, "gifts": 3}


def have():
    """Return True is the player has the Crimbo Shrub in their terrarium, False otherwise."""
    return _have(familiar)


def is_decorated():
    """Return True if the Crimbo Shrub has been decorated today, False otherwise."""
    return get_property("_shrubDecorated", bool)


def check_decorations():
    """Current Crimbo Shrub decorations."""
    return (
        get_property("shrubTopper"),
        get_property("shrubLights"),
        get_property("shrubGarland"),
        get_property("shrubGifts"),
    )


def decorate(topper, lights, garland, gift):
    """Decorate the Crimbo Shrub."""
    if not have():
        raise RuntimeError("need a Crimbo Shrub")
    if is_decorated() and check_decorations() == (topper, lights, garland, gift):
        return
    if is_decorated():
        raise RuntimeError("already been decorated the shrub today")

    choices = (
        topper_choices[topper],
        lights_choices[lights],
        garland_choices[garland],
        gift_choices[gift],
    )
    ash.visit_url(f"inv_use.php?pwd=&which=99&whichitem={decorations.id}")
    ash.visit_url(f"choice.php?whichchoice=999&pwd=&option=1&topper={0}&lights={1}&garland={2}&gift={3}".format(*choices))  # fmt: skip

    if check_decorations() != (topper, lights, garland, gift):
        raise RuntimeError(f"failed to decorate the shrub with {(topper, lights, garland, gift)}")  # fmt: skip



import pymafia.kolmafia as km
from pymafia import ash, datatypes


class Skill:
    id = -1
    name = "none"

    def __init__(self, key=None):
        if key in (None, self.id, self.name):
            return

        id_ = km.SkillDatabase.getSkillId(key) if isinstance(key, str) else key
        name = km.SkillDatabase.getSkillName(id_)

        if name is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        self.id = id_
        self.name = name

    def __str__(self):
        ids = km.SkillDatabase.getSkillIds(self.name, False)
        return f"[{self.id}]{self.name}" if len(ids) > 1 else self.name

    def __repr__(self):
        return f"{type(self).__name__}({str(self)!r})"

    def __hash__(self):
        return hash((self.id, self.name))

    def __eq__(self, other):
        return isinstance(other, type(self)) and (self.id, self.name) == (
            other.id,
            other.name,
        )

    def __bool__(self):
        return (self.id, self.name) != (type(self).id, type(self).name)

    @classmethod
    def all(cls):
        values = km.DataTypes.SKILL_TYPE.allValues()
        return sorted(ash.to_python(values), key=lambda x: x.id)

    @property
    def type_(self):
        return km.SkillDatabase.getSkillTypeName(self.id)

    @property
    def level(self):
        return km.SkillDatabase.getSkillLevel(self.id)

    @property
    def image(self):
        return km.SkillDatabase.getSkillImage(self.id)

    @property
    def traincost(self):
        return km.SkillDatabase.getSkillPurchaseCost(self.id)

    @property
    def class_(self):
        return datatypes.Class(km.SkillDatabase.getSkillCategory(self.id) or None)

    @property
    def libram(self):
        return km.SkillDatabase.isLibramSkill(self.id)

    @property
    def passive(self):
        return km.SkillDatabase.isPassive(self.id)

    @property
    def buff(self):
        return km.SkillDatabase.isBuff(self.id)

    @property
    def combat(self):
        return km.SkillDatabase.isCombat(self.id)

    @property
    def song(self):
        return km.SkillDatabase.isSong(self.id)

    @property
    def expression(self):
        return km.SkillDatabase.isExpression(self.id)

    @property
    def walk(self):
        return km.SkillDatabase.isWalk(self.id)

    @property
    def summon(self):
        return km.SkillDatabase.isSummon(self.id)

    @property
    def permable(self):
        return km.SkillDatabase.isPermable(self.id)

    @property
    def dailylimit(self):
        return km.SkillDatabase.getMaxCasts(self.id)

    @property
    def timescast(self):
        return km.SkillDatabase.getCasts(self.id)



import json
import os
import re
import sys
import urllib
import zipfile

import jnius_config

JAR_LOCATION = "./kolmafia.jar"
JENKINS_JOB_URL = "https://ci.kolmafia.us/job/Kolmafia/lastSuccessfulBuild/"
JAVA_PATTERN = "(net\\/sourceforge\\/kolmafia.*\\/([^\\$]*))\\.class"


def download(location):
    with urllib.request.urlopen(JENKINS_JOB_URL + "/api/json") as response:
        data = json.loads(response.read().decode())
        jar_url = JENKINS_JOB_URL + "artifact/" + data["artifacts"][0]["relativePath"]
        urllib.request.urlretrieve(jar_url, filename=location)


if not os.path.isfile(JAR_LOCATION):
    download(JAR_LOCATION)
jnius_config.set_classpath(JAR_LOCATION)
from jnius import autoclass, cast  # pylint: disable=C0413,E0611,W0611

classes = {}
with zipfile.ZipFile(JAR_LOCATION) as archive:
    for file in archive.filelist:
        filename = file.orig_filename
        match = re.search(JAVA_PATTERN, filename)
        if match:
            classes[match.group(2)] = match.group(1)


def __getattr__(key):
    return autoclass(classes[key]) if key in classes else autoclass(key)


class MafiaError(Exception):
    pass


def tracer(frame, event, arg):
    """Monitor return events for a mafia error.

    See https://stackoverflow.com/questions/59088671/hooking-every-function-call-in-python.
    """
    if event == "call":
        # disable per-line events on the frame to improve performance
        frame.f_trace_lines = False
        return tracer

    elif event == "return":
        KoLmafia = autoclass("net/sourceforge/kolmafia/KoLmafia")
        if not KoLmafia.permitsContinue():
            KoLmafia.forceContinue()
            raise MafiaError(KoLmafia.getLastMessage())


sys.settrace(tracer)



from enum import Enum

import pymafia.kolmafia as km


class Phylum(Enum):
    NONE = "none"
    BEAST = "beast"
    BUG = "bug"
    CONSTELLATION = "constellation"
    CONSTRUCT = "construct"
    DEMON = "demon"
    DUDE = "dude"
    ELEMENTAL = "elemental"
    ELF = "elf"
    FISH = "fish"
    GOBLIN = "goblin"
    HIPPY = "hippy"
    HOBO = "hobo"
    HORROR = "horror"
    HUMANOID = "humanoid"
    MER_KIN = "mer-kin"
    ORC = "orc"
    PENGUIN = "penguin"
    PIRATE = "pirate"
    PLANT = "plant"
    SLIME = "slime"
    UNDEAD = "undead"
    WEIRD = "weird"

    def __str__(self):
        return self.value

    def __bool__(self):
        return self is not self.NONE

    @classmethod
    def _missing_(cls, value):
        if value is None:
            return cls.NONE
        return super()._missing_(value)

    @classmethod
    def all(cls):
        return list(cls)

    @property
    def image(self):
        phylum = km.autoclass(
            "net/sourceforge/kolmafia/persistence/MonsterDatabase$Phylum"
        ).find(self.value)
        if type(self)(phylum.toString()) is self.NONE:
            return ""
        return phylum.getImage()



from .bounty import Bounty
from .class_ import Class
from .coinmaster import Coinmaster
from .effect import Effect, EffectQuality
from .element import Element
from .familiar import Familiar
from .item import CandyType, Item, ItemQuality
from .location import Location
from .monster import Monster
from .phylum import Phylum
from .servant import Servant
from .skill import Skill
from .slot import Slot
from .stat import Stat
from .thrall import Thrall
from .vykea import Vykea



import pymafia.kolmafia as km
from pymafia import ash, datatypes


class Bounty:
    name = "none"

    def __init__(self, key=None):
        if key in (None, self.name):
            return

        bounties = km.BountyDatabase.getMatchingNames(key)

        if len(bounties) != 1:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        canonical = bounties[0]
        self.name = km.BountyDatabase.canonicalToName(canonical)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"{type(self).__name__}({str(self)!r})"

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.name == other.name

    def __bool__(self):
        return self.name != type(self).name

    @classmethod
    def all(cls):
        values = km.DataTypes.BOUNTY_TYPE.allValues()
        return sorted(ash.to_python(values), key=lambda x: x.name)

    @property
    def plural(self):
        return km.BountyDatabase.getPlural(self.name) or ""

    @property
    def type_(self):
        return km.BountyDatabase.getType(self.name) or ""

    @property
    def kol_internal_type(self):
        if self.type_ == "easy":
            return "low"
        if self.type_ == "hard":
            return "high"
        return None

    @property
    def number(self):
        return km.BountyDatabase.getNumber(self.name)

    @property
    def image(self):
        return km.BountyDatabase.getImage(self.name) or ""

    @property
    def monster(self):
        return datatypes.Monster(km.BountyDatabase.getMonster(self.name)) or None

    @property
    def location(self):
        return datatypes.Location(km.BountyDatabase.getLocation(self.name)) or None



from pymafia import ash
from pymafia.combat import Macro
from pymafia.datatypes import Skill
from pymafia.utils import get_property
from pymafia.utils import have as _have
from pymafia.utils import in_choice, in_combat

passive = Skill("Comprehensive Cartography")
skill = Skill("Map the Monsters")


def have():
    """Return True if the player has the Comprehensive Cartography skill, False otherwise."""
    return _have(passive)


def monsters_mapped():
    """Return the number of Map the Monsters skill uses today."""
    return get_property("_monstersMapped", int)


def map_monster(location, monster, macro=Macro()):
    """Map to a monster in a location."""
    if not have():
        raise RuntimeError("need the Comprehensive Cartography skill")
    if monsters_mapped() >= 3:
        raise RuntimeError("already mapped three monsters today")

    if not get_property("mappingMonsters"):
        ash.use_skill(skill)
    ash.visit_url(location.url)
    if not in_choice(1435):
        raise RuntimeError("failed to encounter the Leading Yourself Right to Them noncombat adventure")  # fmt: skip
    ash.visit_url(f"choice.php?pwd=&whichchoice=1435&option=1&heyscriptswhatsupwinkwink={monster.id}")  # fmt: skip
    if not in_combat(monster):
        raise RuntimeError(f"failed to start fight with {monster!r}")
    ash.run_combat(macro)



import collections

import pymafia.kolmafia as km
from pymafia.datatypes import (
    Bounty,
    Class,
    Coinmaster,
    Effect,
    Element,
    Familiar,
    Item,
    Location,
    Monster,
    Phylum,
    Servant,
    Skill,
    Slot,
    Stat,
    Thrall,
    Vykea,
)

SIMPLE_TYPES = {
    km.DataTypes.TYPE_BOOLEAN: bool,
    km.DataTypes.TYPE_INT: int,
    km.DataTypes.TYPE_FLOAT: float,
    km.DataTypes.TYPE_STRING: str,
    km.DataTypes.TYPE_BUFFER: str,
    km.DataTypes.TYPE_ITEM: Item,
    km.DataTypes.TYPE_LOCATION: Location,
    km.DataTypes.TYPE_CLASS: Class,
    km.DataTypes.TYPE_STAT: Stat,
    km.DataTypes.TYPE_SKILL: Skill,
    km.DataTypes.TYPE_EFFECT: Effect,
    km.DataTypes.TYPE_FAMILIAR: Familiar,
    km.DataTypes.TYPE_SLOT: Slot,
    km.DataTypes.TYPE_MONSTER: Monster,
    km.DataTypes.TYPE_ELEMENT: Element,
    km.DataTypes.TYPE_COINMASTER: Coinmaster,
    km.DataTypes.TYPE_PHYLUM: Phylum,
    km.DataTypes.TYPE_BOUNTY: Bounty,
    km.DataTypes.TYPE_THRALL: Thrall,
    km.DataTypes.TYPE_SERVANT: Servant,
    km.DataTypes.TYPE_VYKEA: Vykea,
}

TreeMap = km.autoclass("java.util.TreeMap")
ArrayList = km.autoclass("java.util.ArrayList")
String = km.autoclass("java.lang.String")
ByteArrayInputStream = km.autoclass("java.io.ByteArrayInputStream")


def __getattr__(name):
    return AshFunction(name)


def to_java(obj):
    if isinstance(obj, (type(None), int, float, str)):
        return km.Value(obj)

    if isinstance(obj, tuple(SIMPLE_TYPES.values())):
        parse_value = getattr(km.DataTypes, f"parse{type(obj).__name__}Value")
        return parse_value(str(obj), False)

    if isinstance(obj, collections.abc.Mapping):
        jmap = TreeMap()
        for k, v in obj.items():
            jk = to_java(k)
            jv = to_java(v)
            jmap.put(jk, jv)
        data_type = jmap.getFirstEntry().getValue().getType()
        index_type = jmap.getFirstEntry().getKey().getType()
        aggregate_type = km.AggregateType(data_type, index_type)
        return km.MapValue(aggregate_type, jmap)

    if isinstance(obj, collections.abc.Iterable):
        jlist = ArrayList()
        for item in obj:
            jitem = to_java(item)
            jlist.add(jitem)
        data_type = jlist.get(0).getType()
        size = jlist.size()
        aggregate_type = km.AggregateType(data_type, size)
        return km.ArrayValue(aggregate_type, jlist)

    raise TypeError(f"unsupported type: {type(obj).__name__!r}")


def to_python(obj):
    jtype = obj.getType().getType()
    jname = obj.getType().getName()

    if jtype in [km.DataTypes.TYPE_VOID, km.DataTypes.TYPE_ANY]:
        return None
    if jtype in SIMPLE_TYPES:
        return SIMPLE_TYPES[jtype](obj.toJSON())
    if jtype == km.DataTypes.TYPE_AGGREGATE and isinstance(obj.content, TreeMap):
        return {to_python(x.key): to_python(x.value) for x in obj.content.entrySet()}
    if jtype == km.DataTypes.TYPE_AGGREGATE and isinstance(obj.content, list):
        return [to_python(x) for x in obj.content]

    raise TypeError(f"unsupported type: {jtype!r} ({jname!r})")


def ashref(command=""):
    names = set()
    for func in km.RuntimeLibrary.getFunctions():
        name = func.getName()
        if command.lower() in name:
            names.add(name)
    return sorted(names)


def script(lines, raw=False):
    stream = ByteArrayInputStream(String(lines).getBytes())
    interpreter = km.AshRuntime()
    interpreter.validate(None, stream)
    value = interpreter.execute("main", None)
    return value if raw else to_python(value)


class AshFunction:
    def __init__(self, name):
        self.name = name
        self.func = getattr(km.RuntimeLibrary, self.name)

    def __call__(self, *args, raw=False):
        interpreter = km.AshRuntime()
        jargs = [to_java(arg) for arg in args]
        value = self.func(interpreter, *jargs)
        return value if raw else to_python(value)

    @property
    def signatures(self):
        functions = km.RuntimeLibrary.getFunctions().findFunctions(self.name)
        return [f"{f.getType().toString()} {f.getSignature()}" for f in functions]

    @property
    def wiki(self):
        return f"https://wiki.kolmafia.us/index.php/{self.name}"



from . import (
    barrel_god,
    boxing_daycare,
    cartography,
    cosplay_saber,
    crimbo_shrub,
    god_lobster,
    kramco,
    love_tunnel,
    mumming_trunk,
    pantogram,
    protonic_pack,
    snojo,
    songboom,
    stomping_boots,
    witchess,
)



from pymafia import ash
from pymafia.datatypes import Familiar
from pymafia.utils import get_property
from pymafia.utils import have as _have

familiar = Familiar("Pair of Stomping Boots")


def have():
    """Return True if the player has the Pair of Stomping Boots in their terrarium, False otherwise."""
    return _have(familiar)


def runaways_used():
    """Return the number of free bander runaways used today."""
    return get_property("banderRunaways", int)


def runaways_left():
    """Return the total number of free bander runaways the player can get from their Stomping Boots."""
    return (ash.familiar_weight(familiar) + ash.weight_adjustment()) // 5



import pymafia.kolmafia as km
from pymafia import ash, datatypes

Integer = km.autoclass("java.lang.Integer")


class Monster:
    id = 0
    name = "none"
    monster = None

    def __init__(self, key=None):
        if key in (None, self.id, self.name):
            return

        monster = (
            km.MonsterDatabase.findMonster(key, True)
            if isinstance(key, str)
            else km.MonsterDatabase.findMonsterById(key)
        )

        if monster is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        self.id = monster.getId()
        self.name = monster.getName()
        self.monster = monster

    def __str__(self):
        ids = km.MonsterDatabase.getMonsterIds(self.name, False)
        return f"[{self.id}]{self.name}" if len(ids) > 1 else self.name

    def __repr__(self):
        return f"{type(self).__name__}({str(self)!r})"

    def __hash__(self):
        return hash((self.id, self.name))

    def __eq__(self, other):
        return isinstance(other, type(self)) and (self.id, self.name) == (
            other.id,
            other.name,
        )

    def __bool__(self):
        return (self.id, self.name) != (type(self).id, type(self).name)

    @classmethod
    def all(cls):
        values = km.DataTypes.MONSTER_TYPE.allValues()
        return sorted(ash.to_python(values), key=lambda x: x.id)

    @property
    def base_hp(self):
        return self.monster.getHP() if self else 0

    @property
    def base_attack(self):
        return self.monster.getAttack() if self else 0

    @property
    def raw_hp(self):
        return self.monster.getRawHP() if self else 0

    @property
    def raw_attack(self):
        return self.monster.getRawAttack() if self else 0

    @property
    def raw_defense(self):
        return self.monster.getRawDefense() if self else 0

    @property
    def base_defense(self):
        return self.monster.getDefense() if self else 0

    @property
    def base_initiative(self):
        return self.monster.getInitiative() if self else 0

    @property
    def raw_initiative(self):
        return self.monster.getRawInitiative() if self else 0

    @property
    def attack_element(self):
        return (
            datatypes.Element(self.monster.getAttackElement().toString())
            if self
            else datatypes.Element.NONE
        )

    @property
    def defense_element(self):
        return (
            datatypes.Element(self.monster.getDefenseElement().toString())
            if self
            else datatypes.Element.NONE
        )

    @property
    def physical_resistance(self):
        return self.monster.getPhysicalResistance() if self else 0

    @property
    def min_meat(self):
        return self.monster.getMinMeat() if self else 0

    @property
    def max_meat(self):
        return self.monster.getMaxMeat() if self else 0

    @property
    def min_sprinkles(self):
        return self.monster.getMinSprinkles() if self else 0

    @property
    def max_sprinkles(self):
        return self.monster.getMaxSprinkles() if self else 0

    @property
    def base_mainstat_exp(self):
        return self.monster.getExperience() if self else 0

    @property
    def phylum(self):
        return (
            datatypes.Phylum(self.monster.getPhylum().toString())
            if self
            else datatypes.Phylum.NONE
        )

    @property
    def poison(self):
        if not self:
            return None
        poison_level = self.monster.getPoison()
        if poison_level == Integer.MAX_VALUE:
            return None
        poison_name = km.EffectDatabase.getEffectName(
            km.EffectDatabase.POISON_ID[poison_level]
        )
        return datatypes.Effect(poison_name)

    @property
    def boss(self):
        return self.monster.isBoss() if self else False

    @property
    def copyable(self):
        return self.monster.isNoCopy() if self else False

    @property
    def image(self):
        return self.monster.getImage() if self else ""

    @property
    def images(self):
        return list(self.monster.getImages()) if self else []

    @property
    def random_modifiers(self):
        return list(self.monster.getRandomModifiers()) if self else []

    @property
    def sub_types(self):
        return list(self.monster.getSubTypes()) if self else []

    @property
    def manuel_name(self):
        return self.monster.getManuelName() if self else ""

    @property
    def wiki_name(self):
        return self.monster.getWikiName() if self else ""

    @property
    def attributes(self):
        return self.monster.getAttributes() if self else ""



import re
from html import escape

import pymafia.kolmafia as km
from pymafia import ash
from pymafia.datatypes import Effect, Familiar, Item, Monster, Servant, Skill

HOLIDAY_WANDERERS = {
    "El Dia De Los Muertos Borrachos": [
        Monster("Novia Cadáver"),
        Monster("Novio Cadáver"),
        Monster("Padre Cadáver"),
        Monster("Persona Inocente Cadáver"),
    ],
    "Feast of Boris": [
        Monster("Candied Yam Golem"),
        Monster("Malevolent Tofurkey"),
        Monster("Possessed Can of Cranberry Sauce"),
        Monster("Stuffing Golem"),
    ],
    "Talk Like a Pirate Day": [
        Monster("ambulatory pirate"),
        Monster("migratory pirate"),
        Monster("peripatetic pirate"),
    ],
}

ByteArrayOutputStream = km.autoclass("java.io.ByteArrayOutputStream")
PrintStream = km.autoclass("java.io.PrintStream")


def force_continue():
    km.autoclass("net/sourceforge/kolmafia/KoLmafia").forceContinue()


def launch_gui():
    km.KoLmafia.main(["--GUI"])


def login(username, password=None):
    if password is None:
        password = km.KoLmafia.getSaveState(username)

    request = km.LoginRequest(username, password)
    request.run()
    return request


def abort(message=None):
    km.KoLmafia.updateDisplay(km.KoLConstants.MafiaState.ABORT, message)


def log(message="", html=False):
    message = str(message)
    if not html:
        message = escape(message)

    km.RequestLogger.printLine(message)


def execute(command):
    ostream = km.cast("java.io.OutputStream", ByteArrayOutputStream())
    out = PrintStream(ostream)
    km.RequestLogger.openCustom(out)
    km.KoLmafiaCLI.DEFAULT_SHELL.executeLine(command)
    return ostream.toString()


def get_property(name, t=str):
    if t is bool:
        return km.Preferences.getBoolean(name)
    if t is int:
        return km.Preferences.getInteger(name)
    if t is float:
        return km.Preferences.getFloat(name)
    return t(km.Preferences.getString(name))


def set_property(name, value=""):
    if isinstance(value, bool):
        return km.Preferences.setBoolean(name, value)
    if isinstance(value, int):
        return km.Preferences.setInteger(name, value)
    if isinstance(value, float):
        return km.Preferences.setFloat(name, value)
    return km.Preferences.setString(name, str(value))


def have(thing, quantity=1):
    if isinstance(thing, Effect):
        return ash.have_effect(thing) >= quantity
    if isinstance(thing, Familiar):
        return ash.have_familiar(thing)
    if isinstance(thing, Item):
        return ash.available_amount(thing) >= quantity
    if isinstance(thing, Servant):
        return ash.have_servant(thing)
    if isinstance(thing, Skill):
        return ash.have_skill(thing)
    raise TypeError(f"unexpected type {type(thing).__name__!r}")


def in_choice(choice):
    return ash.handling_choice() and ash.last_choice() == choice


def in_combat(monster=None):
    if ash.current_round() < 1:
        return False
    if monster is None:
        return True

    page = ash.visit_url("fight.php")
    match = re.search("<!-- MONSTERID: (\\d+) -->", page)
    if not match:
        raise RuntimeError("unable to identify monster")
    return int(match.group(1)) == monster.id


def can_kmail():
    return not (
        ash.current_round() > 0  # In a fight
        or ash.handling_choice()  # In a choice
        or ash.fight_follows_choice()  # Was in a choice, gonna be in a fight
        or ash.choice_follows_fight()  # Was in a fight, gonna be in a choice
        or ash.in_multi_fight()  # Was in a fight, gonna be in another fight
    )


def try_use(item):
    if have(item):
        ash.use(1, item)


def get_todays_holiday_wanderers():
    today = ash.holiday()
    if not today:
        return []
    return [mon for holiday in today for mon in HOLIDAY_WANDERERS[holiday.split("/")]]



from enum import Enum


class Element(Enum):
    NONE = "none"
    COLD = "cold"
    HOT = "hot"
    SLEAZE = "sleaze"
    SPOOKY = "spooky"
    STENCH = "stench"
    SLIME = "slime"
    SUPERCOLD = "supercold"

    def __str__(self):
        return self.value

    def __bool__(self):
        return self is not self.NONE

    @classmethod
    def _missing_(cls, value):
        if value is None:
            return cls.NONE
        return super()._missing_(value)

    @classmethod
    def all(cls):
        return list(cls)

    @property
    def image(self):
        if self is self.NONE:
            return "circle.gif"
        if self is self.COLD:
            return "snowflake.gif"
        if self is self.HOT:
            return "fire.gif"
        if self is self.SLEAZE:
            return "wink.gif"
        if self is self.SPOOKY:
            return "skull.gif"
        if self is self.STENCH:
            return "stench.gif"
        # No image for Slime or Supercold in Manuel
        if self is self.SLIME:
            return "circle.gif"
        if self is self.SUPERCOLD:
            return "circle.gif"
        return ""



import pymafia.kolmafia as km
from pymafia import ash, datatypes


class Coinmaster:
    name = "none"
    coinmaster = None

    def __init__(self, key=None):
        if key in (None, self.name):
            return

        coinmaster = km.CoinmasterRegistry.findCoinmaster(key)

        if coinmaster is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        self.name = coinmaster.getMaster()
        self.coinmaster = coinmaster

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"{type(self).__name__}({str(self)!r})"

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.name == other.name

    def __bool__(self):
        return self.name != type(self).name

    @classmethod
    def all(cls):
        values = km.DataTypes.COINMASTER_TYPE.allValues()
        return sorted(ash.to_python(values), key=lambda x: x.name)

    @property
    def token(self):
        return self.coinmaster.getToken() if self else None

    @property
    def item(self):
        if not self:
            return None
        item = self.coinmaster.getItem()
        return None if item is None else datatypes.Item(item.getItemId())

    @property
    def property_(self):
        return self.coinmaster.getProperty() if self else None

    @property
    def available_tokens(self):
        return self.coinmaster.availableTokens() if self else 0

    @property
    def buys(self):
        return self.coinmaster.getSellAction() is not None if self else False

    @property
    def sells(self):
        return self.coinmaster.getBuyAction() is not None if self else False

    @property
    def nickname(self):
        return self.coinmaster.getNickname() if self else None



from enum import Enum


class Stat(Enum):
    NONE = "none"
    MUSCLE = "Muscle"
    MYSTICALITY = "Mysticality"
    MOXIE = "Moxie"
    SUBMUSCLE = "SubMuscle"
    SUBMYSTICALITY = "SubMysticality"
    SUBMOXIE = "SubMoxie"

    def __str__(self):
        return self.value

    def __bool__(self):
        return self is not self.NONE

    @classmethod
    def _missing_(cls, value):
        if value is None:
            return cls.NONE
        return super()._missing_(value)

    @classmethod
    def all(cls):
        return list(cls)



from enum import IntEnum

import pymafia.kolmafia as km
from pymafia import ash


class EffectQuality(IntEnum):
    NONE = -1
    GOOD = 0
    NEUTRAL = 1
    BAD = 2


class Effect:
    id = -1
    name = "none"

    def __init__(self, key=None):
        if key in (None, self.id, self.name):
            return

        id_ = km.EffectDatabase.getEffectId(key) if isinstance(key, str) else key
        name = km.EffectDatabase.getEffectName(id_)

        if name is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        self.id = id_
        self.name = name

    def __str__(self):
        ids = km.EffectDatabase.getEffectIds(self.name, False)
        return f"[{self.id}]{self.name}" if len(ids) > 1 else self.name

    def __repr__(self):
        return f"{type(self).__name__}({str(self)!r})"

    def __hash__(self):
        return hash((self.id, self.name))

    def __eq__(self, other):
        return isinstance(other, type(self)) and (self.id, self.name) == (
            other.id,
            other.name,
        )

    def __bool__(self):
        return (self.id, self.name) != (type(self).id, type(self).name)

    @classmethod
    def all(cls):
        values = km.DataTypes.EFFECT_TYPE.allValues()
        return sorted(ash.to_python(values), key=lambda x: x.id)

    @property
    def default(self):
        return km.EffectDatabase.getDefaultAction(self.id) or ""

    @property
    def quality(self):
        return EffectQuality(km.EffectDatabase.getQuality(self.id))

    @property
    def attributes(self):
        attrs = km.EffectDatabase.getEffectAttributes(self.id)
        return [] if attrs is None else list(attrs)

    @property
    def actions(self):
        return list(km.EffectDatabase.getAllActions(self.id))

    @property
    def note(self):
        return km.EffectDatabase.getActionNote(self.id) or ""

    @property
    def image(self):
        return km.EffectDatabase.getImageName(self.id)

    @property
    def descid(self):
        return km.EffectDatabase.getDescriptionId(self.id) or ""

    @property
    def candy_tier(self):
        return km.CandyDatabase.getEffectTier(self.id)



import pymafia.kolmafia as km
from pymafia import ash, datatypes


class Thrall:
    id = 0
    name = "none"
    data = None

    def __init__(self, key=None):
        if key in (None, self.id, self.name):
            return

        data = (
            km.PastaThrallData.typeToData(key)
            if isinstance(key, str)
            else km.PastaThrallData.idToData(key)
        )

        if data is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        self.id = data[1]
        self.name = data[0]
        self.data = data

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"{type(self).__name__}({str(self)!r})"

    def __hash__(self):
        return hash((self.id, self.name))

    def __eq__(self, other):
        return isinstance(other, type(self)) and (self.id, self.name) == (
            other.id,
            other.name,
        )

    def __bool__(self):
        return (self.id, self.name) != (type(self).id, type(self).name)

    @classmethod
    def all(cls):
        values = km.DataTypes.THRALL_TYPE.allValues()
        return sorted(ash.to_python(values), key=lambda x: x.id)

    @property
    def thrall(self):
        return km.KoLCharacter.findPastaThrall(self.name)

    @property
    def level(self):
        return self.thrall.getLevel() if self else 0

    @property
    def image(self):
        return self.data[6] if self else ""

    @property
    def tinyimage(self):
        return self.data[7] if self else ""

    @property
    def skill(self):
        return datatypes.Skill(self.data[3]) if self else None

    @property
    def current_modifiers(self):
        return self.thrall.getCurrentModifiers() if self else ""



from pymafia import ash
from pymafia.utils import get_property


def have():
    """Return True if the player has the Boxing Daycare open, False otherwise."""
    return get_property("daycareOpen", bool)


def daydream():
    """Have a Boxing Daydream."""
    if not have():
        raise RuntimeError("need access to the Boxing Daycare")
    if get_property("_daycareNap", bool):
        return

    ash.cli_execute("daycare item")


def free_scavenge():
    """Free scavenge for gym equipment."""
    if not have():
        raise RuntimeError("need access to the Boxing Daycare")
    if get_property("_daycareGymScavenges", int) > 0:
        return

    ash.visit_url("place.php?whichplace=town_wrong&action=townwrong_boxingdaycare")
    ash.run_choice(3)  # Enter the Boxing Daycare
    ash.run_choice(2)  # Scavenge for gym equipment
    ash.run_choice(5)  # Return to the Lobby
    ash.run_choice(4)  # Leave

    if get_property("_daycareGymScavenges", int) != 1:
        raise RuntimeError("failed to use the free scavenge")



from enum import Enum

import pymafia.kolmafia as km
from pymafia import ash, datatypes


class ItemQuality(Enum):
    NONE = ""
    CRAPPY = "crappy"
    DECENT = "decent"
    GOOD = "good"
    AWESOME = "awesome"
    EPIC = "EPIC"


class CandyType(Enum):
    NONE = "none"
    UNSPADED = "unspaded"
    SIMPLE = "simple"
    COMPLEX = "complex"


class Item:
    id = -1
    name = "none"

    def __init__(self, key=None):
        if key in (None, self.id, self.name):
            return

        id_ = km.ItemDatabase.getItemId(key) if isinstance(key, str) else key
        name = km.ItemDatabase.getItemDataName(id_)

        if name is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        self.id = id_
        self.name = name

    def __str__(self):
        ids = km.ItemDatabase.getItemIds(self.name, 1, False)
        return f"[{self.id}]{self.name}" if len(ids) > 1 else self.name

    def __repr__(self):
        return f"{type(self).__name__}({str(self)!r})"

    def __hash__(self):
        return hash((self.id, self.name))

    def __eq__(self, other):
        return isinstance(other, type(self)) and (self.id, self.name) == (
            other.id,
            other.name,
        )

    def __bool__(self):
        return (self.id, self.name) != (type(self).id, type(self).name)

    @classmethod
    def all(cls):
        values = km.DataTypes.ITEM_TYPE.allValues()
        return ash.to_python(values)

    @property
    def plural(self):
        """Return the plural of the Item. If the official plural is not known, return the name of the Item with an "s" appended."""
        return km.ItemDatabase.getPluralName(self.id)

    @property
    def descid(self):
        """Return the descid of the Item. This is the identifier used to see the description of the Item."""
        return km.ItemDatabase.getDescriptionId(self.id)

    @property
    def image(self):
        """Return the filename of the image associated with the Item."""
        return km.ItemDatabase.getImage(self.id)

    @property
    def smallimage(self):
        """Returns the filename of the small image associated with the Item. For items with an image that is usually larger than 30x30, returns their 30x30 equivalent.

        For example, "folders" from the "over-the-shoulder Folder Holder" will normally return a 100x100 image but a 30x30 image here.
        """
        return km.ItemDatabase.getSmallImage(self.id)

    @property
    def levelreq(self):
        """Return the level requirement for consuming or equipping the Item."""
        return km.ConsumablesDatabase.getLevelReqByName(self.name) or 0

    @property
    def quality(self):
        """Return the quality of the Item if it is a consumable."""
        return ItemQuality(km.ConsumablesDatabase.getQuality(self.name))

    @property
    def adventures(self):
        """Return the range of adventures gained from consuming the Item. The string will either contain the adventures for invariant gains, or a hyphen-separated minimum and maximum for variant gains."""
        return km.ConsumablesDatabase.getAdvRangeByName(self.name)

    @property
    def muscle(self):
        """Return the range of muscle substats gained from consuming the Item. The string will either contain the substats for invariant gains, or a hyphen-separated minimum and maximum for variant gains. Note that substat gains can be negative."""
        return km.ConsumablesDatabase.getMuscleByName(self.name)

    @property
    def mysticality(self):
        """Return the range of mysticality substats gained from consuming the Item. The string will either contain the substats for invariant gains, or a hyphen-separated minimum and maximum for variant gains. Note that substat gains can be negative."""
        return km.ConsumablesDatabase.getMysticalityByName(self.name)

    @property
    def moxie(self):
        """Return the range of moxie substats gained from consuming the Item. The string will either contain the substats for invariant gains, or a hyphen-separated minimum and maximum for variant gains. Note that substat gains can be negative."""
        return km.ConsumablesDatabase.getMoxieByName(self.name)

    @property
    def fullness(self):
        """Return the stomach size of Item. If the Item is not edible, return 0."""
        return km.ConsumablesDatabase.getFullness(self.name)

    @property
    def inebriety(self):
        """Return the liver size of Item. If the Item is not drinkable, return 0."""
        return km.ConsumablesDatabase.getInebriety(self.name)

    @property
    def spleen(self):
        """Return the spleen size of Item. If the Item is not chewable, return 0."""
        return km.ConsumablesDatabase.getSpleenHit(self.name)

    @property
    def minhp(self):
        """Return the minimum HP restored by consuming this Item."""
        return km.RestoresDatabase.getHPMin(self.name)

    @property
    def maxhp(self):
        """Return the maximum HP restored by consuming this Item."""
        return km.RestoresDatabase.getHPMax(self.name)

    @property
    def minmp(self):
        """Return the minimum MP restored by consuming this Item."""
        return km.RestoresDatabase.getMPMin(self.name)

    @property
    def maxmp(self):
        """Return the maximum MP restored by consuming this Item."""
        return km.RestoresDatabase.getMPMax(self.name)

    @property
    def dailyusesleft(self):
        """Return the number of daily uses remaining for this Item."""
        return km.UseItemRequest.maximumUses(self.id)

    @property
    def notes(self):
        """Return any notes that exist for the Item. Examples of (comma-separated) contents are:

        - The name and duration of any effects granted by consumption, if applicable.
        - Items dropped when the item is consumed, if applicable.
        - Tags relevant to game mechanics (such as "MARTINI", "BEER" and "SAUCY")
        - "Unspaded"
        """
        return km.ConsumablesDatabase.getNotes(self.name) or ""

    @property
    def quest(self):
        """Return True if the Item is a quest item, else False."""
        return km.ItemDatabase.isQuestItem(self.id)

    @property
    def gift(self):
        """Return True if the Item is a gift item, else False."""
        return km.ItemDatabase.isGiftItem(self.id)

    @property
    def tradeable(self):
        """Return True if the Item is tradeable, else False."""
        return km.ItemDatabase.isTradeable(self.id)

    @property
    def discardable(self):
        """Return True if the Item is discardable, else False."""
        return km.ItemDatabase.isDiscardable(self.id)

    @property
    def combat(self):
        """Return True if the Item is usable in combat, else False. This returns True whether the Item is consumed by being used or not."""
        return km.ItemDatabase.getAttribute(
            self.id, km.ItemDatabase.ATTR_COMBAT | km.ItemDatabase.ATTR_COMBAT_REUSABLE
        )

    @property
    def combat_reusable(self):
        """Return True if the Item is usable in combat and is not consumed when doing so, else False."""
        return km.ItemDatabase.getAttribute(
            self.id, km.ItemDatabase.ATTR_COMBAT_REUSABLE
        )

    @property
    def usable(self):
        """Return True if the Item is usable, else False. This returns True whether the Item is consumed by being used or not."""
        return km.ItemDatabase.isUsable(self.id)

    @property
    def reusable(self):
        """Return True if the Item is usable and is not consumed when doing so, else False."""
        return km.ItemDatabase.getConsumptionType(
            self.id
        ) == km.KoLConstants.INFINITE_USES or km.ItemDatabase.getAttribute(
            self.id, km.ItemDatabase.ATTR_REUSABLE
        )

    @property
    def multi(self):
        """Return True if the Item is multiusable, else False."""
        return km.ItemDatabase.isMultiUsable(self.id)

    @property
    def fancy(self):
        """Return True if the Item is a "fancy" ingredient, else False."""
        return km.ItemDatabase.isFancyItem(self.id)

    @property
    def pasteable(self):
        """Return True if the Item is a meatpasting ingredient, else False."""
        return km.ItemDatabase.isPasteable(self.id)

    @property
    def smithable(self):
        """Return True if the Item is a meatsmithing ingredient, else False."""
        return km.ItemDatabase.isSmithable(self.id)

    @property
    def cookable(self):
        """Return True if the Item is a cooking ingredient, else False."""
        return km.ItemDatabase.isCookable(self.id)

    @property
    def mixable(self):
        """Return True if the Item is a cocktailcrafting ingredient, else False."""
        return km.ItemDatabase.isMixable(self.id)

    @property
    def candy(self):
        """Return True if the Item is a candy, else False."""
        return km.ItemDatabase.isCandyItem(self.id)

    @property
    def candy_type(self):
        """Return the candy type of the Item."""
        return CandyType(km.CandyDatabase.getCandyType(self.id))

    @property
    def chocolate(self):
        """Return True if the Item is a chocolate, else False."""
        return km.ItemDatabase.isChocolateItem(self.id)

    @property
    def seller(self):
        """Return which Coinmaster sells this Item, if any."""
        data = km.CoinmasterRegistry.findSeller(self.id)
        return None if data is None else datatypes.Coinmaster(data.getMaster())

    @property
    def buyer(self):
        """Return which Coinmaster buys this Item, if any."""
        data = km.CoinmasterRegistry.findBuyer(self.id)
        return None if data is None else datatypes.Coinmaster(data.getMaster())

    @property
    def name_length(self):
        """Return the length of the Item's display name."""
        return km.ItemDatabase.getNameLength(self.id)

    @property
    def noob_skill(self):
        """Return the noob Skill granted by absorbing this Item."""
        return datatypes.Skill(km.ItemDatabase.getNoobSkillId(self.id)) or None

    @property
    def tcrs_name(self):
        """Return the name of the Item as it appears in your current Two Crazy Random Summer run. If you are not in a TCRS run, the regular Item name is returned."""
        return km.TCRSDatabase.getTCRSName(self.id) or ""



class Macro:
    def __init__(self, commands=None):
        self.commands = [] if commands is None else commands

    def __str__(self):
        return ";".join(self.commands) + ";" if self.commands else ""

    def __repr__(self):
        return f"{type(self).__name__}({self.commands!r})"

    def __add__(self, other):
        return type(self)(self.commands + other.commands)

    def step(self, steps):
        if isinstance(steps, str):
            return self + type(self)([steps])
        return self + steps

    def abort(self, message=""):
        return self.step(f'abort "{message}"')

    def attack(self):
        return self.step("attack")

    def repeat(self):
        return self.step("repeat")

    def runaway(self):
        return self.step("runaway")

    def skill(self, skill):
        return self.step(f"skill {skill.id}")

    def try_skill(self, skill):
        return self.step(f"if hasskill {skill.id}").skill(skill).step("endif")

    def item(self, item):
        return self.step(f"use {item.id}")

    def if_monster(self, monster, macro):
        return self.step(f"if monsterid {monster.id}").step(macro).step("endif")

    def check_monster(self, monsters):
        monsters = monsters if isinstance(monsters, list) else [monsters]
        predicate = " || ".join([f"monsterid {x.id}" for x in monsters])
        return self.step(f"if !({predicate})").abort("unexpected monster").step("endif")



from pymafia import ash
from pymafia.combat import Macro
from pymafia.datatypes import Familiar, Item
from pymafia.utils import get_property
from pymafia.utils import have as _have

familiar = Familiar("God Lobster")

reward_choices = {"regalia": 1, "blessing": 2, "experience": 3}
regalia = [
    Item("God Lobster's Scepter"),
    Item("God Lobster's Ring"),
    Item("God Lobster's Rod"),
    Item("God Lobster's Robe"),
    Item("God Lobster's Crown"),
]


def have():
    """Return True if the player has the God Lobster in their terrarium, False otherwise."""
    return _have(familiar)


def fights_today():
    """Return the number of God Lobster fights used today."""
    return get_property("godLobsterFights", int)


def fights_left():
    """Return the number of God Lobster fights remaining today."""
    return 3 - fights_today()


def fight(reward, macro=Macro()):
    """Fight the God Lobster and choose a reward."""
    if not have():
        raise RuntimeError("need a God Lobster")
    if fights_left() < 1:
        raise RuntimeError("out of God Lobster fights")

    choice = reward_choices[reward]
    ash.use_familiar(familiar)
    initial_fights = fights_today()
    ash.visit_url("main.php?fightgodlobster=1")
    ash.run_combat(macro)
    ash.run_choice(choice)

    if fights_today() != initial_fights + 1:
        raise RuntimeError("failed to fight the God Lobster")



from pymafia import ash
from pymafia.datatypes import Item
from pymafia.utils import get_property
from pymafia.utils import have as _have

item = Item("SongBoom™ BoomBox")

song_keywords = {
    "Eye of the Giger": "spooky",
    "Food Vibrations": "food",
    "Remainin' Alive": "dr",
    "These Fists Were Made for Punchin'": "damage",
    "Total Eclipse of Your Meat": "meat",
    "Silence": "off",
    None: "off",
}


def have():
    """Return True if the player has the SongBoom™ BoomBox available, False otherwise."""
    return _have(item)


def song():
    """Return the current SongBoom™ Boombox song."""
    return get_property("boomBoxSong") or None


def song_changes_left():
    """Return the number of SongBoom™ Boombox song changes left today."""
    return get_property("_boomBoxSongsLeft", int)


def set_song(new_song):
    """Change the SongBoom™ Boombox song."""
    if not have():
        raise RuntimeError("need a SongBoom™ BoomBox")
    if song() == new_song:
        return
    if song_changes_left() < 1:
        raise RuntimeError("out of song changes")

    ash.cli_execute(f"boombox {song_keywords[new_song]}")


def drop_progress():
    """Return the progress to next SongBoom™ Boombox drop (e.g. gathered meat-clip)."""
    return get_property("_boomBoxFights", int)



import pymafia.kolmafia as km
from pymafia import ash, datatypes


class Location:
    id = -1
    name = "none"
    adventure = None

    def __init__(self, key=None):
        if key in (None, self.id, self.name):
            return

        adventure = (
            km.AdventureDatabase.getAdventure(key)
            if isinstance(key, str)
            else km.AdventureDatabase.getAdventureByURL(
                f"adventure.php?snarfblat={key}"
            )
        )

        if adventure is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        self.id = adventure.getSnarfblat()
        self.name = adventure.getAdventureName()
        self.adventure = adventure

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"{type(self).__name__}({str(self)!r})"

    def __hash__(self):
        return hash((self.id, self.name))

    def __eq__(self, other):
        return isinstance(other, type(self)) and (self.id, self.name) == (
            other.id,
            other.name,
        )

    def __bool__(self):
        return (self.id, self.name) != (type(self).id, type(self).name)

    @classmethod
    def all(cls):
        values = km.DataTypes.LOCATION_TYPE.allValues()
        return sorted(ash.to_python(values), key=lambda x: x.id)

    @property
    def url(self):
        return f"adventure.php?snarfblat={self.id}"

    @property
    def nocombats(self):
        return self.adventure.isNonCombatsOnly() if self else False

    @property
    def combat_percent(self):
        if not self:
            return 0
        area = self.adventure.getAreaSummary()
        return 0 if area is None else area.areaCombatPercent()

    @property
    def zone(self):
        return self.adventure.getZone() if self else ""

    @property
    def parent(self):
        return self.adventure.getParentZone() if self else ""

    @property
    def parentdesc(self):
        return self.adventure.getParentZoneDescription() if self else ""

    @property
    def environment(self):
        return self.adventure.getEnvironment() if self else ""

    @property
    def bounty(self):
        if not self:
            return None
        bounty = km.AdventureDatabase.getBounty(self.adventure)
        return None if bounty is None else datatypes.Bounty(bounty.getName())

    @property
    def combat_queue(self):
        if not self:
            return []
        zone_queue = km.AdventureQueueDatabase.getZoneQueue(self.adventure)
        return [] if zone_queue is None else list(zone_queue)

    @property
    def noncombat_queue(self):
        if not self:
            return []
        zone_queue = km.AdventureQueueDatabase.getZoneNoncombatQueue(self.adventure)
        return [] if zone_queue is None else list(zone_queue)

    @property
    def turns_spent(self):
        return km.AdventureSpentDatabase.getTurns(self.adventure, True) if self else 0

    @property
    def kisses(self):
        return km.FightRequest.dreadKisses(self.adventure) if self else 0

    @property
    def recommended_stat(self):
        return self.adventure.getRecommendedStat() if self else 0

    @property
    def water_level(self):
        return (
            self.adventure.getWaterLevel()
            if self and km.KoLCharacter.inRaincore()
            else 0
        )

    @property
    def wanderers(self):
        return self.adventure.hasWanderers() if self else False

    @property
    def fire_level(self):
        return (
            km.WildfireCampRequest.getFireLevel(self.adventure)
            if self and km.KoLCharacter.inFirecore()
            else 0
        )



from enum import Enum


class Slot(Enum):
    NONE = "none"
    HAT = "hat"
    WEAPON = "weapon"
    HOLSTER = "holster"
    OFFHAND = "off-hand"
    BACK = "back"
    SHIRT = "shirt"
    PANTS = "pants"
    ACC1 = "acc1"
    ACC2 = "acc2"
    ACC3 = "acc3"
    FAMILIAR = "familiar"
    CROWNOFTHRONES = "crown-of-thrones"
    STICKER1 = "sticker1"
    STICKER2 = "sticker2"
    STICKER3 = "sticker3"
    CARDSLEEVE = "card-sleeve"
    FOLDER1 = "folder1"
    FOLDER2 = "folder2"
    FOLDER3 = "folder3"
    FOLDER4 = "folder4"
    FOLDER5 = "folder5"
    BUDDYBJORN = "buddy-bjorn"
    BOOTSKIN = "bootskin"
    BOOTSPUR = "bootspur"
    FAKEHAND = "fakehand"

    def __str__(self):
        return self.value

    def __bool__(self):
        return self is not self.NONE

    @classmethod
    def _missing_(cls, value):
        if value is None:
            return cls.NONE
        return super()._missing_(value)

    @classmethod
    def all(cls):
        return list(cls)



from pymafia import ash
from pymafia.datatypes import Item
from pymafia.utils import get_property
from pymafia.utils import have as _have

item = Item("mumming trunk")

costume_choices = {
    "meat": 0,
    "mp": 1,
    "musc": 2,
    "item": 3,
    "myst": 4,
    "hp": 5,
    "mox": 6,
}


def have():
    """Return True if the player has the mumming trunk available, False otherwise."""
    return _have(item)


def costumes_used():
    """Return a list of the costumes applied today."""
    uses = [int(x) for x in get_property("_mummeryUses").split(",") if x]
    return [name for name, choice in costume_choices.items() if choice in uses]


def apply_costume(costume):
    """Dress up the player's current familiar with a costume."""
    if not have():
        raise RuntimeError("need a mumming trunk")
    if costume in costumes_used():
        raise RuntimeError(f"already applied the {costume!r} costume today")
    if not ash.my_familiar():
        raise RuntimeError("need to have a familiar to put a costume on")

    choice = costume_choices[costume]
    ash.cli_execute(f"mummery {choice}")



from pymafia import ash
from pymafia.combat import Macro
from pymafia.datatypes import Item, Monster
from pymafia.utils import get_property, in_choice, in_combat

item = Item("Witchess Set")

pieces = [
    Monster("Witchess Pawn"),
    Monster("Witchess Knight"),
    Monster("Witchess Bishop"),
    Monster("Witchess Rook"),
    Monster("Witchess Ox"),
    Monster("Witchess King"),
    Monster("Witchess Witch"),
    Monster("Witchess Queen"),
]


def have():
    """Return True if the player has the Witchess Set in their campground, False otherwise."""
    return item in ash.get_campground()


def fights_today():
    """Return the number of Witchess fights used today."""
    return get_property("_witchessFights", int)


def fights_left():
    """Return the number of Witchess fights left today."""
    return 5 - fights_today()


def fight(piece, macro=Macro()):
    """Fight a Witchess piece."""
    if not have():
        raise RuntimeError("need a Witchess Set installed")
    if fights_left() < 1:
        raise RuntimeError("out of Witchess fights")
    if piece not in pieces:
        raise ValueError(f"unknown piece: {piece!r}")

    ash.visit_url("campground.php?action=witchess")
    if not in_choice(1181):
        raise RuntimeError("failed to open Witchess")
    ash.run_choice(1)
    if not in_choice(1182):
        raise RuntimeError("failed to visit shrink ray")
    ash.visit_url(
        f"choice.php?option=1&pwd={ash.my_hash()}&whichchoice=1182&piece={piece.id}",
        False,
    )
    if not in_combat(piece):
        raise RuntimeError("failed to start fight")
    ash.run_combat(macro)



import pymafia.kolmafia as km
from pymafia import ash, datatypes


class Class:
    id = -1
    name = "none"
    ascension_class = None

    def __init__(self, key=None):
        if key in (None, self.id, self.name):
            return

        ascension_class = km.AscensionClass.find(key)

        if ascension_class is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        self.id = ascension_class.getId()
        self.name = ascension_class.getName()
        self.ascension_class = ascension_class

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"{type(self).__name__}({str(self)!r})"

    def __hash__(self):
        return hash((self.id, self.name))

    def __eq__(self, other):
        return isinstance(other, type(self)) and (self.id, self.name) == (
            other.id,
            other.name,
        )

    def __bool__(self):
        return (self.id, self.name) != (type(self).id, type(self).name)

    @classmethod
    def all(cls):
        values = km.DataTypes.CLASS_TYPE.allValues()
        return sorted(ash.to_python(values), key=lambda x: x.id)

    @property
    def primestat(self):
        if not self:
            return datatypes.Stat.NONE
        prime_index = self.ascension_class.getPrimeStatIndex()
        stat_name = km.AdventureResult.STAT_NAMES[prime_index]
        return datatypes.Stat[stat_name.upper()]



from pymafia import ash
from pymafia.datatypes import Item
from pymafia.utils import get_property
from pymafia.utils import have as _have

item = Item("Kramco Sausage-o-Matic™")


def have():
    """Return True if the player has the Kramco Sausage-o-Matic™ available, False otherwise."""
    return _have(item)


def is_fight_ready():
    """Return True if the maximum turns between sausage goblins has been reached, False otherwise."""
    if not have():
        return False

    total_fought = get_property("_sausageFights", int)
    last_turn = get_property("_lastSausageMonsterTurn", int)
    ready_on = last_turn + 4 + 3 * total_fought + max(0, total_fought - 5) ** 3
    return ash.total_turns_played() >= ready_on



from pymafia import ash
from pymafia.datatypes import Location
from pymafia.utils import get_property

location = Location("The X-32-F Combat Training Snowman")

setting_choices = {
    "MUSCLE": 1,
    "MYSTICALITY": 2,
    "MOXIE": 3,
    "TOURNAMENT": 4,
}


def have():
    """Return True if the player has The Snojo available, False otherwise"""
    return get_property("snojoAvailable", bool)


def setting():
    """Return the current snojo setting."""
    return get_property("snojoSetting")


def free_fights_today():
    """Return the number of free snojo fights used today."""
    return get_property("_snojoFreeFights", int)


def free_fights_left():
    """Return the number of free snojo fights left today."""
    return 10 - free_fights_today()


def change_setting(new_setting):
    """Change the snojo setting."""
    if not have():
        raise RuntimeError("need access to The Snojo")
    if new_setting == setting():
        return

    choice = setting_choices[new_setting]
    ash.visit_url("place.php?whichplace=snojo&action=snojo_controller")
    ash.run_choice(choice)

    if setting() != new_setting:
        raise RuntimeError("failed to change snojo setting")



from pymafia import ash
from pymafia.datatypes import Item
from pymafia.utils import get_property
from pymafia.utils import have as _have

item = Item("Fourth of May Cosplay Saber")
upgrade_choices = {"mp": 1, "ml": 2, "resistance": 3, "familiar": 4}


def have():
    """Return True if the player has the Fourth of May Cosplay Saber available, False otherwise."""
    return _have(item)


def current_upgrade():
    """Return the current Fourth of May Cosplay Saber upgrade."""
    mod = get_property("_saberMod", int)
    for name, choice in upgrade_choices.items():
        if choice == mod:
            return name
    return None


def is_upgraded():
    """Return True if the Fourth of May Cosplay Saber has been upgraded today, False otherwise."""
    return current_upgrade() is not None


def upgrade(new_upgrade):
    """Upgrade the Fourth of May Cosplay Saber."""
    if not have():
        raise RuntimeError("need a Fourth of May Cosplay Saber")
    if is_upgraded() and current_upgrade() == new_upgrade:
        return
    if is_upgraded():
        raise RuntimeError("already upgraded the saber today")
    if new_upgrade not in upgrade_choices:
        raise ValueError(f"unknown upgrade: {new_upgrade!r}")

    ash.cli_execute(f"saber {new_upgrade}")



import pymafia.kolmafia as km
from pymafia import ash


class Servant:
    id = 0
    name = "none"
    data = None

    def __init__(self, key):
        if key in (None, self.id, self.name):
            return

        data = (
            km.EdServantData.typeToData(key)
            if isinstance(key, str)
            else km.EdServantData.idToData(key)
        )

        if data is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        self.id = data[2]
        self.name = data[0]
        self.data = data

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"{type(self).__name__}({str(self)!r})"

    def __hash__(self):
        return hash((self.id, self.name))

    def __eq__(self, other):
        return isinstance(other, type(self)) and (self.id, self.name) == (
            other.id,
            other.name,
        )

    def __bool__(self):
        return (self.id, self.name) != (type(self).id, type(self).name)

    @classmethod
    def all(cls):
        values = km.DataTypes.SERVANT_TYPE.allValues()
        return sorted(ash.to_python(values), key=lambda x: x.id)

    @property
    def servant(self):
        return km.EdServantData.findEdServant(self.name)

    @property
    def level(self):
        return 0 if self.servant is None else self.servant.getLevel()

    @property
    def experience(self):
        return 0 if self.servant is None else self.servant.getExperience()

    @property
    def image(self):
        return self.data[3] if self else ""

    @property
    def level1_ability(self):
        return self.data[4] if self else ""

    @property
    def level7_ability(self):
        return self.data[5] if self else ""

    @property
    def level14_ability(self):
        return self.data[6] if self else ""

    @property
    def level21_ability(self):
        return self.data[7] if self else ""



from enum import IntEnum

import pymafia.kolmafia as km
from pymafia import ash, datatypes


class VykeaType(IntEnum):
    NONE = km.VYKEACompanionData.NONE
    BOOKSHELF = km.VYKEACompanionData.BOOKSHELF
    DRESSER = km.VYKEACompanionData.DRESSER
    CEILING_FAN = km.VYKEACompanionData.CEILING_FAN
    COUCH = km.VYKEACompanionData.COUCH
    LAMP = km.VYKEACompanionData.LAMP
    DISHRACK = km.VYKEACompanionData.DISHRACK


class Vykea:
    companion = km.VYKEACompanionData.NO_COMPANION

    def __init__(self, key=None):
        if key in (None, "none"):
            return

        companion = km.VYKEACompanionData.fromString(key)

        if companion is None:
            raise ValueError(f"{type(self).__name__} {key!r} not found")

        self.companion = companion

    def __str__(self):
        return self.companion.toString()

    def __repr__(self):
        return f"{type(self).__name__}({str(self)!r})"

    def __hash__(self):
        return hash((self.type_, self.rune, self.level))

    def __eq__(self, other):
        return isinstance(other, type(self)) and (
            self.type_,
            self.rune,
            self.level,
        ) == (other.type_, other.rune, other.level)

    def __bool__(self):
        return self.companion != type(self).companion

    @classmethod
    def all(cls):
        values = km.DataTypes.VYKEA_TYPE.allValues()
        return ash.to_python(values)

    @property
    def name(self):
        return self.companion.getName()

    @property
    def type_(self):
        return VykeaType(self.companion.getType())

    @property
    def rune(self):
        item_id = self.companion.getRune().getItemId()
        return None if item_id == datatypes.Item.id else datatypes.Item(item_id)

    @property
    def level(self):
        return self.companion.getLevel()

    @property
    def image(self):
        return self.companion.getImage()

    @property
    def modifiers(self):
        return self.companion.getModifiers()

    @property
    def attack_element(self):
        return datatypes.Element(self.companion.getAttackElement().toString())



from pymafia import ash
from pymafia.datatypes import Item, Location
from pymafia.utils import get_property
from pymafia.utils import have as _have

item = Item("protonic accelerator pack")


def have():
    """Return True if the player has the protonic accelerator pack available, False otherwise."""
    return _have(item)


def ghost_location():
    """Return the current location of the protonic ghost."""
    return get_property("ghostLocation", Location)


def streams_crossed():
    """Return True if the player has crossed streams today, False otherwise."""
    return get_property("_streamsCrossed", bool)


def cross_streams():
    """Cross streams with the target in preference "streamCrossDefaultTarget"."""
    if not have():
        raise RuntimeError("need a protonic accelerator pack")
    if streams_crossed():
        raise RuntimeError("already crossed streams today")

    ash.cli_execute("crossstreams")
